from fastapi import params
import lark
import pysnooper
from lark import Tree, Transformer, Token
from typing import Any, Dict, List, Tuple, Union, Optional, Callable
import json
import plotly.graph_objects as go


def create_parser(start_rule: str = 'network') -> lark.Lark:
    """
    Creates a Lark parser for neural network configuration files.

    Args:
        start_rule (str): The starting rule for the parser. Defaults to 'network'.

    Returns:
        lark.Lark: A Lark parser object configured with the defined grammar.
    """

    grammar = r"""

        VARIABLE: /[a-zA-Z_][a-zA-Z0-9_]*/
        STRING: "\"" /[^"]+/ "\"" | "\'" /[^']+/ "\'"
        %ignore /\#[^\n]*/  // Ignore line comments              
        // Import common tokens from Lark
        %import common.NEWLINE -> _NL        // Newline characters
        %import common.CNAME -> NAME
        %import common.SIGNED_NUMBER -> NUMBER
        %import common.WS_INLINE             // Inline whitespace
        %import common.INT
        %import common.CNAME -> NAME         // Names/identifiers
        %import common.FLOAT  
        %import common.TUPLE                                                        // Floating poNUMBER numbers
        %import common.WS                    // All whitespace
        %ignore WS   

        ?start: network | layer | research   // Allow parsing either networks or research files

        // File type rules
        neural_file: network
        nr_file: network
        rnr_file: research

        // Parameter & Properties
        named_params: named_param ("," named_param)*
        activation_param: "activation" "=" STRING
        ordered_params: value ("," value)* 
        ?value: STRING -> string_value | number | tuple_ | BOOL  
        tuple_: "("  number  ","  number  ")"  
        number: NUMBER  
        number1: INT
        BOOL: "true" | "false"
        explicit_tuple: "(" value ("," value)+ ")"

        // name_param rules
        bool_value: BOOL  // Example: true or false
        named_return_sequences: "return_sequences" "=" bool_value
        named_units: "units" "=" number  // Example: units=64
        named_activation: "activation" "=" STRING
        named_size: NAME ":" explicit_tuple  
        named_filters: "filters" "=" NUMBER  // Example: filters=32
        named_strides: "strides" "=" value  // Example: strides=(1, 1) or strides=1
        named_padding: "padding" "=" STRING | "padding" ":" STRING  // Example: padding="same" or padding="valid"
        named_dilation_rate: "dilation_rate" "=" value  // Example: dilation_rate=(2, 2) or dilation_rate=2
        named_groups: "groups" "=" NUMBER  // Example: groups=32
        named_channels: "channels" "=" NUMBER  // Example: channels=3
        named_num_heads: "num_heads" "=" NUMBER  // Example: num_heads=8
        named_ff_dim: "ff_dim" "=" NUMBER  // Example: ff_dim=512
        named_input_dim: "input_dim" "=" NUMBER  // Example: input_dim=1000
        named_output_dim: "output_dim" "=" NUMBER  // Example: output_dim=128
        named_rate: "rate" "=" FLOAT  // Example: rate=0.5
        named_dropout: "dropout" "=" FLOAT  // Example: dropout=0.2
        named_axis: "axis" "=" NUMBER  // Example: axis=1
        named_momentum: "momentum" "=" FLOAT  // Example: momentum=0.9
        named_epsilon: "epsilon" "=" FLOAT  // Example: epsilon=1e-05
        named_center: "center" "=" BOOL  // Example: center=true
        named_scale: "scale" "=" BOOL  // Example: scale=true
        named_beta_initializer: "beta_initializer" "=" STRING  // Example: beta_initializer="zeros"
        named_gamma_initializer: "gamma_initializer" "=" STRING  // Example: gamma_initializer="ones"
        named_moving_mean_initializer: "moving_mean_initializer" "=" STRING  // Example: moving_mean_initializer="zeros"
        named_moving_variance_initializer: "moving_variance_initializer" "=" STRING  // Example: moving_variance_initializer="ones"
        named_training: "training" "=" BOOL  // Example: training=true
        named_trainable: "trainable" "=" BOOL  // Example: trainable=false
        named_use_bias: "use_bias" "=" BOOL  // Example: use_bias=true
        named_kernel_initializer: "kernel_initializer" "=" STRING  // Example: kernel_initializer="glorot_uniform"
        named_bias_initializer: "bias_initializer" "=" STRING  // Example: bias_initializer="zeros"
        named_kernel_regularizer: "kernel_regularizer" "=" STRING  // Example: kernel_regularizer="l2"
        named_bias_regularizer: "bias_regularizer" "=" STRING  // Example: bias_regularizer="l1"
        named_activity_regularizer: "activity_regularizer" "=" STRING  // Example: activity_regularizer="l1_l2"
        named_kernel_constraint: "kernel_constraint" "=" STRING  // Example: kernel_constraint="max_norm"
        named_kernel_size: "kernel_size" "=" value // Example:
        named_bias_constraint: "bias_constraint" "=" STRING  // Example: bias_constraint="min_max_norm"
        named_return_state: "return_state" "=" BOOL  // Example: return_state=true
        named_go_backwards: "go_backwards" "=" BOOL  // Example: go_backwards=false
        named_stateful: "stateful" "=" BOOL  // Example: stateful=true
        named_time_major: "time_major" "=" BOOL  // Example: time_major=false
        named_unroll: "unroll" "=" BOOL  // Example: unroll=true
        named_input_shape: "input_shape" "=" value  // Example: input_shape=(28, 28, 1)
        named_batch_input_shape: "batch_input_shape" "=" value  // Example: batch_input_shape=(None, 32, 32, 3)
        named_dtype: "dtype" "=" STRING  // Example: dtype="float32"
        named_name: "name" "=" STRING  // Example: name="my_layer"
        named_weights: "weights" "=" value  // Example: weights=[...]
        named_embeddings_initializer: "embeddings_initializer" "=" STRING  // Example: embeddings_initializer="uniform"
        named_mask_zero: "mask_zero" "=" BOOL  // Example: mask_zero=true
        named_input_length: "input_length" "=" NUMBER  // Example: input_length=100
        named_embeddings_regularizer: "embeddings_regularizer" "=" STRING  // Example: embeddings_regularizer="l1"
        named_embeddings_constraint: "embeddings_constraint" "=" value // Example: embeddings_constraint="non_neg"
        named_num_layers: "num_layers" "=" NUMBER // Example: num_layers=2
        named_bidirectional: "bidirectional" "=" BOOL // Example: bidirectional=true
        named_merge_mode: "merge_mode" "=" STRING // Example: merge_mode="concat"
        named_recurrent_dropout: "recurrent_dropout" "=" FLOAT // Example: recurrent_dropout=0.1
        named_noise_shape: "noise_shape" "=" value // Example: noise_shape=(3,)
        named_seed: "seed" "=" NUMBER // Example: seed=42
        named_target_shape: "target_shape" "=" value // Example: target_shape=(64, 64)
        named_data_format: "data_format" "=" STRING // Example: data_format="channels_first"
        named_interpolation: "interpolation" "=" STRING // Example: interpolation="nearest"
        named_crop_to_aspect_ratio: "crop_to_aspect_ratio" "=" BOOL // Example: crop_to_aspect_ratio=true
        named_mask_value: "mask_value" "=" NUMBER // Example: mask_value=0
        named_return_attention_scores: "return_attention_scores" "=" BOOL // Example: return_attention_scores=true
        named_causal: "causal" "=" BOOL // Example: causal=false
        named_use_scale: "use_scale" "=" BOOL // Example: use_scale=true
        named_key_dim: "key_dim" "=" NUMBER // Example: key_dim=64
        named_value_dim: "value_dim" "=" NUMBER 
        named_output_shape: "output_shape" "=" value 
        named_arguments: "arguments" "=" value 
        named_initializer: "initializer" "=" STRING 
        named_regularizer: "regularizer" "=" STRING 
        named_constraint: "constraint" "=" STRING
        named_l1: "l1" "=" FLOAT  // Example: l1=0.01
        named_l2: "l2" "=" FLOAT  // Example: l2=0.001
        named_l1_l2: "l1_l2" "=" tuple_  // Example: l1_l2=(0.01, 0.001)
        named_int: NAME "=" INT | NAME ":" INT
        named_string: NAME "=" STRING | NAME ":" STRING
        named_float: NAME "=" FLOAT | NAME ":" FLOAT
        simple_number: number1
        rate: "rate" ":" FLOAT
        simple_float: FLOAT
        named_clipvalue: "clipvalue=" FLOAT
        named_clipnorm: "clipnorm=" FLOAT
        ?named_param: ( rate | named_clipvalue | named_clipnorm | simple_float | explicit_tuple | simple_number| named_units | pool_size | named_kernel_size | named_size | named_activation | named_filters | named_strides | named_padding | named_dilation_rate | named_groups | named_data_format | named_channels | named_return_sequences | named_num_heads | named_ff_dim | named_input_dim | named_output_dim | named_rate | named_dropout | named_axis | named_momentum | named_epsilon | named_center | named_scale | named_beta_initializer | named_gamma_initializer | named_moving_mean_initializer | named_moving_variance_initializer | named_training | named_trainable | named_use_bias | named_kernel_initializer | named_bias_initializer | named_kernel_regularizer | named_bias_regularizer | named_activity_regularizer | named_kernel_constraint | named_bias_constraint | named_return_state | named_go_backwards | named_stateful | named_time_major | named_unroll | named_input_shape | named_batch_input_shape | named_dtype | named_name | named_weights | named_embeddings_initializer | named_mask_zero | named_input_length | named_embeddings_regularizer | named_embeddings_constraint | named_num_layers | named_bidirectional | named_merge_mode | named_recurrent_dropout | named_noise_shape | named_seed | named_target_shape | named_interpolation | named_crop_to_aspect_ratio | named_mask_value | named_return_attention_scores | named_causal | named_use_scale | named_key_dim | named_value_dim | named_output_shape | named_arguments | named_initializer | named_regularizer | named_constraint | named_l1 | named_l2 | named_l1_l2 | named_int | named_float | NAME "=" value )
        
        // Layer parameter styles
        ?param_style1: named_params | (value ("," (value | named_param))* )

        // Top-level network definition - defines the structure of an entire neural network
        network: "network" NAME "{" input_layer layers loss optimizer [training_config] [execution_config] "}" 

        // Configuration can be either training-related or execution-related
        config: training_config | execution_config

        // Input layer definition - specifies the shape of input data
        input_layer: "input" ":" shape ("," shape)*
        shape: "(" [number_or_none ("," number_or_none)* [","]] ")"
        number_or_none: number | "NONE" | "None"

        // Layers section - contains all layer definitions separated by newlines
        layers: "layers" ":" _NL* (layer _NL*)*

        
        // Lambda functions
        lambda_: "Lambda("  STRING  ")"

        // Wrapper Layer Functions 
        wrapper: "TimeDistributed" "(" layer ["," named_params] ")"

        
        // Basic layer types & group
        ?basic: conv | pooling | dropout | flatten | dense | output
        dropout: "Dropout(" named_params ")" 
        dense: "Dense" "(" dense_params ")"
        dense_params: (NUMBER ("," (NUMBER | STRING | named_param))* ) | named_params
        // dense_ordered_params: value ("," value)*  // Allow any valid 'value'
        flatten: "Flatten" "(" [named_params] ")"
        // Recurrent layers section - includes all RNN variants
        ?recurrent: rnn | bidirectional_rnn | conv_rnn | rnn_cell  
        bidirectional_rnn: "Bidirectional(" rnn "," named_params ")" 
        rnn: simple_rnn | lstm | gru
        simple_rnn: "SimpleRNN(" named_params ")"
        lstm: "LSTM(" named_params ")"
        gru: "GRU(" named_params ")"
        

        // Regularization layers group
        regularization: spatial_dropout1d | spatial_dropout2d | spatial_dropout3d | activity_regularization | l1 | l2 | l1_l2 
        l1: "L1(" named_params ")"
        l2: "L2(" named_params ")"
        l1_l2: "L1L2(" named_params ")" 


        // Output layer 
        output: "Output(" named_params ")"

        // Convolutional layers 
        conv: conv1d | conv2d | conv3d | conv_transpose | depthwise_conv2d | separable_conv2d
        conv1d: "Conv1D" "(" param_style1 ")"
        conv2d: "Conv2D" "(" param_style1 ")"
        conv3d: "Conv3D" "(" param_style1 ")"
        conv_transpose: conv1d_transpose | conv2d_transpose | conv3d_transpose
        conv1d_transpose: "Conv1DTranspose" "(" param_style1 ")"
        conv2d_transpose: "Conv2DTranspose" "(" param_style1 ")"
        conv3d_transpose: "Conv3DTranspose" "(" param_style1 ")"
        depthwise_conv2d: "DepthwiseConv2D" "(" param_style1 ")"
        separable_conv2d: "SeparableConv2D" "(" param_style1 ")"

        // Pooling layer parameters
        pooling: max_pooling | average_pooling | global_pooling | adaptive_pooling
        max_pooling: max_pooling1d | max_pooling2d | max_pooling3d
        max_pooling1d: "MaxPooling1D" "(" named_params ")"
        max_pooling2d: "MaxPooling2D" "(" param_style1 ")"
        max_pooling3d: "MaxPooling3D" "(" named_params ")"
        pool_size : "pool_size" "=" value
        average_pooling: average_pooling1d | average_pooling2d | average_pooling3d
        average_pooling1d: "AveragePooling1D(" named_params ")"
        average_pooling2d: "AveragePooling2D(" named_params ")"
        average_pooling3d: "AveragePooling3D(" named_params ")"
        global_pooling: global_max_pooling | global_average_pooling
        global_max_pooling: global_max_pooling1d | global_max_pooling2d | global_max_pooling3d
        global_max_pooling1d: "GlobalMaxPooling1D(" named_params ")"
        global_max_pooling2d: "GlobalMaxPooling2D(" named_params ")"
        global_max_pooling3d: "GlobalMaxPooling3D(" named_params ")"
        global_average_pooling: global_average_pooling1d | global_average_pooling2d | global_average_pooling3d
        global_average_pooling1d: "GlobalAveragePooling1D(" named_params ")"
        global_average_pooling2d: "GlobalAveragePooling2D(" named_params ")"
        global_average_pooling3d: "GlobalAveragePooling3D(" named_params ")"
        adaptive_pooling: adaptive_max_pooling | adaptive_average_pooling
        adaptive_max_pooling: adaptive_max_pooling1d | adaptive_max_pooling2d | adaptive_max_pooling3d
        adaptive_max_pooling1d: "AdaptiveMaxPooling1D(" named_params ")"
        adaptive_max_pooling2d: "AdaptiveMaxPooling2D(" named_params ")"
        adaptive_max_pooling3d: "AdaptiveMaxPooling3D(" named_params ")"
        adaptive_average_pooling: adaptive_average_pooling1d | adaptive_average_pooling2d | adaptive_average_pooling3d
        adaptive_average_pooling1d: "AdaptiveAveragePooling1D(" named_params ")"
        adaptive_average_pooling2d: "AdaptiveAveragePooling2D(" named_params ")"
        adaptive_average_pooling3d: "AdaptiveAveragePooling3D(" named_params ")"

        // Normalization layers
        ?norm_layer: batch_norm | layer_norm | instance_norm | group_norm
        batch_norm: "BatchNormalization" "(" [named_params] ")"
        layer_norm: "LayerNormalization" "(" [named_params] ")"
        instance_norm: "InstanceNormalization" "(" [named_params] ")"
        group_norm: "GroupNormalization" "(" [named_params] ")"
        
        conv_rnn: conv_lstm | conv_gru
        conv_lstm: "ConvLSTM2D(" named_params ")"
        conv_gru: "ConvGRU2D(" named_params ")"

        rnn_cell: simple_rnn_cell | lstm_cell | gru_cell
        simple_rnn_cell: "SimpleRNNCell(" named_params ")"
        lstm_cell: "LSTMCell(" named_params ")"
        gru_cell: "GRUCell(" named_params ")"

        
        // Dropout wrapper layers for RNNs
        dropout_wrapper_layer: simple_rnn_dropout | gru_dropout | lstm_dropout
        simple_rnn_dropout: "SimpleRNNDropoutWrapper" "(" named_params ")"
        gru_dropout: "GRUDropoutWrapper" "(" named_params ")"
        lstm_dropout: "LSTMDropoutWrapper" "(" named_params ")"
        bidirectional_rnn_layer: bidirectional_simple_rnn_layer | bidirectional_lstm_layer | bidirectional_gru_layer
        bidirectional_simple_rnn_layer: "Bidirectional(SimpleRNN(" named_params "))"
        bidirectional_lstm_layer: "Bidirectional(LSTM(" named_params "))"
        bidirectional_gru_layer: "Bidirectional(GRU(" named_params "))"
        conv_rnn_layer: conv_lstm_layer | conv_gru_layer
        conv_lstm_layer: "ConvLSTM2D(" named_params ")"  // Add 2D for clarity
        conv_gru_layer: "ConvGRU2D(" named_params ")"  // Add 2D for clarity
        rnn_cell_layer: simple_rnn_cell_layer | lstm_cell_layer | gru_cell_layer
        simple_rnn_cell_layer: "SimpleRNNCell(" named_params ")"
        lstm_cell_layer: "LSTMCell(" named_params ")"
        gru_cell_layer: "GRUCell(" named_params ")"

        // Advanced layers group
        ?advanced: ( attention | transformer | residual | inception | capsule | squeeze_excitation | graph | embedding | quantum | dynamic )
        attention: "Attention" "(" [named_params] ")"

        // Transformers
        transformer: "Transformer" "(" [named_params] ")" | "TransformerEncoder" "(" [named_params] ")" | "TransformerDecoder" "(" [named_params] ")"
        
        residual: "ResidualConnection" "(" [named_params] ")"
        inception: "Inception" "(" [named_params] ")"
        capsule: "CapsuleLayer" "(" [named_params] ")"
        squeeze_excitation: "SqueezeExcitation" "(" [named_params] ")"
        graph: graph_conv | graph_attention
        graph_conv: "GraphConv" "(" [named_params] ")"
        graph_attention: "GraphAttention" "(" [named_params] ")"
        embedding: "Embedding" "(" [named_params] ")"
        quantum: "QuantumLayer" "(" [named_params] ")"
        dynamic: "DynamicLayer" "(" [named_params] ")"

        merge: add | subtract | multiply | average | maximum | concatenate | dot
        add: "Add(" named_params ")"
        subtract: "Subtract(" named_params ")"
        multiply: "Multiply(" named_params ")"
        average: "Average(" named_params ")"
        maximum: "Maximum(" named_params ")"
        concatenate: "Concatenate(" named_params ")"
        dot: "Dot(" named_params ")"

        noise: gaussian_noise | gaussian_dropout | alpha_dropout
        gaussian_noise: "GaussianNoise(" named_params ")"
        gaussian_dropout: "GaussianDropout(" named_params ")"
        alpha_dropout: "AlphaDropout(" named_params ")"

        spatial_dropout1d: "SpatialDropout1D(" named_params ")"
        spatial_dropout2d: "SpatialDropout2D(" named_params ")"
        spatial_dropout3d: "SpatialDropout3D(" named_params ")"
        activity_regularization: "ActivityRegularization(" named_params ")"

        

        // Activation functions
        activation: activation_with_params | activation_without_params
        activation_with_params: "Activation(" STRING "," named_params ")"  
        activation_without_params: "Activation(" STRING ")" 

        // Training configuration block
        training_config: "train" "{" training_params "}"
        training_params: (epochs_param | batch_size_param | optimizer_param | search_method_param)*
        epochs_param: "epochs:" INT
        batch_size_param: "batch_size:" values_list
        values_list: "[" value ("," value)* "]" | value ("," value)*
        optimizer_param: "optimizer:" named_optimizer 
        named_optimizer : "named_optimizer(" learning_rate_param ")"
        learning_rate_param: "learning_rate=" FLOAT
        search_method_param: "search_method:" STRING


        // Loss and optimizer specifications
        loss: "loss" ":" (NAME | STRING) ["(" named_params ")"]
        optimizer: "optimizer:" (NAME | STRING) ["(" named_params ")"] 
        schedule: NAME "(" params ")"
        params: [value ("," value)*]

        // Execution environment configuration
        execution_config: "execution" "{" device_param "}"
        device_param: "device:" STRING
        // Research-specific configurations
        research: "research" NAME? "{" [research_params] "}"
        research_params: (metrics | references)*
        // Metrics tracking
        metrics: "metrics" "{" [accuracy_param] [metrics_loss_param] [precision_param] [recall_param] "}"
        accuracy_param: "accuracy:" FLOAT
        metrics_loss_param: "loss:" FLOAT
        precision_param: "precision:" FLOAT
        recall_param: "recall:" FLOAT
        // Paper references
        references: "references" "{" paper_param+ "}"
        paper_param: "paper:" STRING

        // Custom Shape Propagation
        custom_shape: "CustomShape" "(" NAME "," explicit_tuple ")"

        // Math Expressions
        math_expr: term (("+"|"-") term)*
        term: factor (("*"|"/") factor)*
        factor: NUMBER | VARIABLE | "(" math_expr ")" | function_call
        function_call: NAME "(" math_expr ("," math_expr)* ")"

        // Automatic Hyperparameters Optimization
        hpo_expr: "HPO(" (hpo_choice | hpo_range | hpo_log_range) ")"
        hpo_choice: "choice(" value ("," value)* ")"
        hpo_range: "range(" number "," number ("," "step="number)? ")"
        hpo_log_range: "log_range(" number "," number ")"


        // Neural Architecture Search
        layer_choice: "HPO(choice(" layer ("," layer)* "))"

        // All possible layer types that can be used in the network
        ?layer: (basic | recurrent | advanced | activation | merge | noise | norm_layer | regularization | custom | wrapper | lambda_ )  
        
        // Custom layer must come after specific layer rules to avoid conflicts
        custom: NAME "(" named_params ")"

    """
    return  lark.Lark(grammar, start=[start_rule], parser='lalr', lexer='contextual', cache=False)

network_parser = create_parser('network')
layer_parser = create_parser('layer')
research_parser = create_parser('research')

class ModelTransformer(lark.Transformer):
    """
    Transforms the parsed tree NUMBER to a structured dictionary representing the neural network model.
    """
    def layer(self, items):
        return self.visit(items[0])

        
    def wrapper(self, items):
        wrapper_type = items[0]
        layer = items[1]
        params = items[2]
        # Merge layer params with wrapper params
        layer['params'].update(params)
        return {'type': f"{wrapper_type}({layer['type']})", 'params': layer['params']}

    ### Basic Layers & Properties ###################

    def input_layer(self, items):
        for item in items:
            shapes = self._extract_value(item)
        return {'type': 'Input', 'shape': shapes}
    
    def layers(self, items):
        return items

    def flatten(self, items):
        if items:
            params = self._extract_value(items[0])
        else:
            params = None
        return {'type': 'Flatten', 'params': params}

    def dropout(self, items):
        params = self._extract_value(items[0])
        return {'type': 'Dropout', 'params': params}

    def output(self, items):
        return {'type': 'Output', 'params': items[0]}
    
    def regularization(self, items):  
        return {'type': items[0].data.capitalize(), 'params': items[0].children[0]}
    
    def execution_config(self, items):
        params = self._extract_value(items[0])
        return {'type': 'execution_config', 'params':params}

    # In the ModelTransformer class:
    def dense(self, items):
        param_nodes = items[0].children
        param_dict = {}
        ordered_params = []
        named_params = {}

        for child in param_nodes:
            param = self._extract_value(child)
            if isinstance(param, dict):  # Named parameter
                named_params.update(param)
            else:  # Ordered parameter
                ordered_params.append(param)

        # Assign ordered params to 'units' and 'activation'
        if ordered_params:
            param_dict['units'] = ordered_params[0]
            if len(ordered_params) > 1:
                param_dict['activation'] = ordered_params[1]
        
        # Merge named parameters
        param_dict.update(named_params)
        
        return {"type": "Dense", "params": param_dict}
    
    ### Convolutional Layers ####################

    def conv(self, items):
        return items[0]

    def conv1d(self, items):
        return {'type': 'Conv1D', 'params': items[0]}

    def conv2d(self, items):
        param_style = items[0]
        raw_params = self._extract_value(param_style)
        ordered_params = []
        named_params = {}

        # Separate ordered and named parameters
        if isinstance(raw_params, list):
            for param in raw_params:
                if isinstance(param, dict):
                    named_params.update(param)
                else:
                    ordered_params.append(param)
        elif isinstance(raw_params, dict):
            named_params = raw_params
        else:
            ordered_params.append(raw_params)

        param_dict = {}
        # Assign ordered parameters to known fields
        if ordered_params:
            if len(ordered_params) >= 1:
                param_dict['filters'] = ordered_params[0]
            if len(ordered_params) >= 2:
                kernel_size = ordered_params[1]
                if isinstance(kernel_size, (list, tuple)):
                    param_dict['kernel_size'] = tuple(kernel_size)
                else:
                    param_dict['kernel_size'] = kernel_size
            if len(ordered_params) >= 3:
                param_dict['activation'] = ordered_params[2]

        # Merge named parameters (override ordered if needed)
        param_dict.update(named_params)

        return {'type': 'Conv2D', 'params': param_dict}
    def conv3d(self, items):
        return {'type': 'Conv3D', 'params': items[0]}
    
    def conv1d_transpose(self, items):
        return {'type': 'Conv1DTranspose', 'params': items[0]}

    def conv2d_transpose(self, items):
        return {'type': 'Conv2DTranspose', 'params': items[0]}

    def conv3d_transpose(self, items):
        return {'type': 'Conv3DTranspose', 'params': items[0]}

    def depthwise_conv2d(self, items):
        return {'type': 'DepthwiseConv2D', 'params': items[0]}

    def separable_conv2d(self, items):
        return {'type': 'SeparableConv2D', 'params': items[0]}
    
    def graph_conv(self, items):
        params = self._extract_value(items[0]) if items else None
        return {'type': 'GraphConv', 'params': params}
    
    ### Loss Parameters ###

    def loss(self, items):
        return items[0].value.strip('"')


    ### Optimization ##############

    def optimizer(self, items):
        name = str(items[0].value)
        # Strip surrounding quotes if present
        if (name.startswith('"') and name.endswith('"')) or (name.startswith("'") and name.endswith("'")):
            name = name[1:-1]
        params = self._extract_value(items[1]) if len(items) > 1 else {}
        
        # if name == "SGD" and "learning_rate" not in params:
           # params["learning_rate"] = 0.01

        return {"type": name, "params": params}

    def schedule(self, items):
        return {
            "type": items[0].value,
            "args": [self._extract_value(x) for x in items[1].children]
        }
    
    ### Training Configurations ############################

    def training_config(self, items):
        for item in items:
            params = self._extract_value(item) if items else None
        return params
    
    def training_params(self, items):
        params = {}
        for item in items:
            if isinstance(item, Tree):
                item = self._extract_value(item)
            elif isinstance(item, dict):
                item = self._extract_value(item)
                params.update(item)
        return params
    
    def epochs_param(self, items):
        return {'epochs': self._extract_value(items[0])}

    def batch_size_param(self, items):
        return {'batch_size': self._extract_value(items[0])}

    def values_list(self, items):
        for x in items:
            params = self._extract_value(x)
        return params
    
    def optimizer_param(self, items):
        return {'optimizer': self._extract_value(items[0])}
    
    def named_optimizer(self, items):
        return {'named_optimizer': self._extract_value(items[0])}

    def learning_rate_param(self,items):
        return {'learning_rate': self._extract_value(items[0])}
    
    def shape(self, items):
        return tuple(items)
    
    ### Wrapper Layers ###

    def wrapper(self, items):
        layer = items[0]
        params = items[1] if len(items) > 1 else {}
        layer['params'].update(params)
        return {'type': f"TimeDistributed({layer['type']})", 'params': layer['params']}

    
    ### Pooling Layers #############################

    def pooling(self, items):
        # Simply return the child rule's result (like max_pooling2d's output)
        return items[0]
    
    def max_pooling(self, items):
        return self._extract_value(items[0])

    def pool_size(self, items):
        value = self._extract_value(items[0])
        return {'pool_size': value}

    def max_pooling1d(self, items):
        param_nodes = items[0].children
        param_dict = {}
        params = [self._extract_value(child) for child in param_nodes]

        if all(isinstance(p, dict) for p in params):
            for p in params:
                param_dict.update(p)
        else:
            if len(params) >= 1:
                param_dict["pool_size"] = params[0]
            if len(params) >= 2:
                param_dict["strides"] = params[1]
            if len(params) >= 3:
                param_dict["padding"] = params[2]

        return {'type': 'MaxPooling1D', 'params': param_dict}

    @pysnooper.snoop()
    def max_pooling2d(self, items):
        # Extract the parsed parameters using _extract_value
        param_style = self._extract_value(items[0])
        params = {}
        
        if isinstance(param_style, list):
            # Extract ordered parameters
            ordered_params = [p for p in param_style if not isinstance(p, dict)]
            if ordered_params:
                params['pool_size'] = ordered_params[0]
            if len(ordered_params) > 1:
                params['strides'] = ordered_params[1]
            if len(ordered_params) > 2:
                params['padding'] = ordered_params[2]
            
            # Extract named parameters
            for item in param_style:
                if isinstance(item, dict):
                    params.update(item)
        elif isinstance(param_style, dict):
            # Use named parameters directly
            params = param_style.copy()
        
        return {'type': 'MaxPooling2D', 'params': params}

    def max_pooling3d(self, items):
        param_nodes = items[0].children
        param_dict = {}
        params = [self._extract_value(child) for child in param_nodes]

        if all(isinstance(p, dict) for p in params):
            for p in params:
                param_dict.update(p)
        else:
            if len(params) >= 1:
                param_dict["pool_size"] = params[0]
            if len(params) >= 2:
                param_dict["strides"] = params[1]
            if len(params) >= 3:
                param_dict["padding"] = params[2]

        return {"type": "MaxPooling3D", "params": param_dict}
    
    def average_pooling1d(self, items):
        return {'type': 'AveragePooling1D', 'params': items[0]}

    def average_pooling2d(self, items):
        return {'type': 'AveragePooling2D', 'params': items[0]}

    def average_pooling3d(self, items):
        return {'type': 'AveragePooling3D', 'params': items[0]}

    def global_max_pooling1d(self, items):
        return {'type': 'GlobalMaxPooling1D', 'params': items[0]}

    def global_max_pooling2d(self, items):
        return {'type': 'GlobalMaxPooling2D', 'params': items[0]}

    def global_max_pooling3d(self, items):
        return {'type': 'GlobalMaxPooling3D', 'params': items[0]} 

    def global_average_pooling1d(self, items):
        return {'type': 'GlobalAveragePooling1D', 'params': items[0]}

    def global_average_pooling2d(self, items):
        return {'type': 'GlobalAveragePooling2D', 'params': items[0]}

    def global_average_pooling3d(self, items):
        return {'type': 'GlobalAveragePooling3D', 'params': items[0]}
    
    def adaptive_max_pooling1d(self, items):
        return {'type': 'AdaptiveMaxPooling1D', 'params': items[0]}

    def adaptive_max_pooling2d(self, items):
        return {'type': 'AdaptiveMaxPooling2D', 'params': items[0]}

    def adaptive_max_pooling3d(self, items):
        return {'type': 'AdaptiveMaxPooling3D', 'params': items[0]}
    
    def adaptive_average_pooling1d(self, items):
        return {'type': 'AdaptiveAveragePooling1D', 'params': items[0]}

    def adaptive_average_pooling2d(self, items):
        return {'type': 'AdaptiveAveragePooling2D', 'params': items[0]}

    def adaptive_average_pooling3d(self, items):
        return {'type': 'AdaptiveAveragePooling3D', 'params': items[0]}

    ### End Basic Layers & Properties #########################

    def batch_norm(self, items):
        params = self._extract_value(items[0]) if items else None
        return {'type': 'BatchNormalization', 'params': params}

    def layer_norm(self, items):
        params = self._extract_value(items[0]) if items else None
        return {'type': 'LayerNormalization', 'params': params}

    def instance_norm(self, items):
        params = self._extract_value(items[0]) if items else None
        return {'type': 'InstanceNormalization', 'params': params}
    def group_norm(self, items):
        params = self._extract_value(items[0]) if items else None
        return {'type': 'GroupNormalization', 'params': params}

    def lstm(self, items):
        return {'type': 'LSTM', 'params': items[0]}

    def gru(self, items):
        return {'type': 'GRU', 'params': items[0]}
    
    ### Recurrent Layers ########################################

    def simple_rnn(self, items):
        return {'type': 'SimpleRNN', 'params': items[0]}
    
    def conv_lstm(self, items):
        return {'type': 'ConvLSTM2D', 'params': items[0]}

    def conv_gru(self, items):
        return {'type': 'ConvGRU2D', 'params': items[0]}

    def bidirectional_rnn(self, items):
        rnn_layer = items[0]
        bidirectional_params = items[1]
        rnn_layer['params'].update(bidirectional_params)  # Merge params
        return {'type': f"Bidirectional({rnn_layer['type']})", 'params': rnn_layer['params']}

    def cudnn_gru_layer(self, items):  # No such thing as CuDNNGRU in PyTorch
        return {'type': 'GRU', 'params': items[0]}

    def bidirectional_simple_rnn_layer(self, items):
        return {'type': 'Bidirectional(SimpleRNN)', 'params': items[0]}

    def bidirectional_lstm_layer(self, items):
        return {'type': 'Bidirectional(LSTM)', 'params': items[0]}

    def bidirectional_gru_layer(self, items):
        return {'type': 'Bidirectional(GRU)', 'params': items[0]}

    def conv_lstm_layer(self, items):
        return {'type': 'ConvLSTM2D', 'params': items[0]}

    def conv_gru_layer(self, items):
        return {'type': 'ConvGRU2D', 'params': items[0]}

    def rnn_cell_layer(self, items):
        return {'type': 'RNNCell', 'params': items[0]}
    
    def simple_rnn_cell(self, items):
        return {'type': 'SimpleRNNCell', 'params': items[0]}

    def lstm_cell(self, items):
        return {'type': 'LSTMCell', 'params': items[0]}

    def gru_cell(self, items):
        return {'type': 'GRUCell', 'params': items[0]}

    def simple_rnn_dropout(self, items):
        return {"type": "SimpleRNNDropoutWrapper", 'params': items[0]}

    def gru_dropout(self, items):
        return {"type": "GRUDropoutWrapper", 'params': items[0]}

    def lstm_dropout(self, items):
        return {"type": "LSTMDropoutWrapper", 'params': items[0]}

    ### Everything Research ##################

    def research(self, items):
        name = None
        params = {}
        
        # Check if the first item is a NAME token (research name)
        if items and isinstance(items[0], Token) and items[0].type == 'NAME':
            name = self._extract_value(items[0])
            # Remaining items are research_params
            if len(items) > 1:
                params = self._extract_value(items[1])
        else:
            # No name, all items are research_params
            if items:
                params = self._extract_value(items[0])
        
        return {'type': 'Research', 'name': name, 'params': params}
    
    def research_params(self, items):
        params = {}
        for item in items:
            if isinstance(item, Tree):
                item = self._extract_value(item)
            elif isinstance(item, dict):
                item = self._extract_value(item)
                params.update(item)
        return params
    
    def metrics(self, items):
        # If no metric items were provided, return an empty dictionary.
        if not items:
            return {'metrics': {}}
        result = {}
        for item in items:
            # Skip any None values
            if item is None:
                continue
            # If the item is already a dictionary (as produced by individual metric rules), update the result.
            if isinstance(item, dict):
                result.update(item)
            else:
                # Otherwise, attempt to extract a string value and split it by colon
                val = self._extract_value(item)
                if val and ':' in val:
                    key, v = str(val).split(':', 1)
                    try:
                        result[key.strip()] = float(v.strip())
                    except ValueError:
                        result[key.strip()] = v.strip()
        return {'metrics': result}

    def accuracy_param(self, items):
        return {'accuracy': self._extract_value(items[0])}

    def precision_param(self, items):
        return {'precision': self._extract_value(items[0])}

    def recall_param(self, items):
        return {'recall': self._extract_value(items[0])}
    
    def references(self, items):
        # If no metric items were provided, return an empty dictionary.
        if not items:
            return {'references': {}}
        result = {}
        for item in items:
            # Skip any None values
            if item is None:
                continue
            # If the item is already a dictionary (as produced by individual references rules), update the result.
            if isinstance(item, dict):
                result.update(item)
            else:
                # Otherwise, attempt to extract a string value and split it by colon
                val = self._extract_value(item)
                if val and ':' in val:
                    key, v = str(val).split(':', 1)
                    try:
                        result[key.strip()] = float(v.strip())
                    except ValueError:
                        result[key.strip()] = v.strip()
        return {'references': result}
    
    def paper_param(self, items):
        return self._extract_value(items[0])


    ### NETWORK ACTIVATION ##############################

    def network(self, items):
        name = str(items[0].value)
        input_layer_config = items[1]
        layers_config = items[2]
        loss_config = items[3]
        optimizer_config = items[4]
        training_config = next((item for item in items[5:] if isinstance(item, dict)), None)
        execution_config = next((item for item in items[5:] if isinstance(item, dict) and 'params' in item and 'device' in item['params']), {'params': {'device': 'auto'}})['params']

        # Ensure output_layer exists
        output_layer = next((layer for layer in reversed(items[2]) 
                        if layer['type'] == 'Output'), None)
        
        output_shape = {}
        if output_layer is None:
            output_shape is None
        elif output_layer is not None:
            output_shape = output_layer.get('params', {}).get('units')
        
        

        return {
            'type': 'model',
            'name': name,
            'input': input_layer_config,
            'layers': layers_config,
            'output_layer': output_layer,
            'output_shape': output_shape,
            'loss': loss_config,
            'optimizer': optimizer_config,
            'training_config': training_config,
            'execution_config': execution_config
        }


    ### Parameters  #######################################################

    def named_params(self, items):
        params = {}
        for item in items:
            if isinstance(item, Tree):
                item = self._extract_value(item)
            elif isinstance(item, dict):
                item = self._extract_value(item)
                params.update(item)
        return params
    
    def _extract_value(self, item):  # Helper function to extract values from tokens and tuples
        if isinstance(item, Token):
            if item.type == 'NAME':
                return item.value
            if item.type in ('INT', 'FLOAT', 'NUMBER', 'SIGNED_NUMBER'):
                try:
                    return int(item.value)
                except ValueError:
                    return float(item.value)
            elif item.type == 'BOOL':
                return item.value.lower() == 'true'
            elif item.type == 'STRING':
                return item.value.strip('"')
            elif item.type == 'WS_INLINE':
                return item.value.strip()
        elif isinstance(item, Tree) and item.data == 'number_or_none':
            child = item.children[0]
            if isinstance(child, Token) and child.value.upper() in ('NONE', 'None'):
                return None
            else:
                return self._extract_value(child)
        # Handle single-value trees (string_value, number, bool_value)
        elif isinstance(item, Tree):
            if item.data == 'string_value':
                return self._extract_value(item.children[0])
            elif item.data == 'number':
                return self._extract_value(item.children[0])
            elif item.data == 'bool_value':
                return self._extract_value(item.children[0])
            elif item.data in ('tuple_', 'explicit_tuple'):
                return tuple(self._extract_value(child) for child in item.children)
            else:
                # Handle lists of parameters (not key-value pairs)
                extracted = [self._extract_value(child) for child in item.children]
                # If any element is a dict, treat as mixed params
                if any(isinstance(e, dict) for e in extracted):
                    return extracted
                # If even number of children, attempt key-value pairing
                if len(item.children) % 2 == 0:
                    try:
                        return {self._extract_value(k): self._extract_value(v) 
                                for k, v in zip(item.children[::2], item.children[1::2])}
                    except TypeError:
                        # Fallback to list if pairing fails
                        return extracted
                else:
                    return extracted
        # Handle nested structures
        elif isinstance(item, list):
            return [self._extract_value(elem) for elem in item]
        elif isinstance(item, dict):
            return {k: self._extract_value(v) for k, v in item.items()}
        return item

    def named_float(self, items):
        # items = [NAME, '='|':', FLOAT] 
        return {items[0].value: self._extract_value(items[1])}

    def named_int(self, items):
        # items = [NAME, '=', INT]
        return {items[0].value: self._extract_value(items[1])}

    def named_param(self, items):
        # Generic NAME "=" value rule (e.g., activation="relu")
        # items = [NAME, '=', value]
        return {items[0].value: self._extract_value(items[1])}
    
    def named_string(self, items):
        # items = [NAME, '='|':', STRING]
        return {items[0].value: self._extract_value(items[1])}
    
    def number(self, items):
        return self._extract_value(items[0])
    
    def rate(self, items):
        params = self._extract_value(items[0])
        return {'type': 'rate', 'params': params}
    
    def simple_float(self, items):
        params = self._extract_value(items[0])
        return {'rate': params}
    
    def number_or_none(self, items):
        if not items:
            # Instead of raising an error, return None if no items are provided.
            return None
        value = self._extract_value(items[0])
        if value == "None":
            return None
        try:
            # Convert to int if there is no decimal point; otherwise, convert to float
            return int(value) if '.' not in str(value) else float(value)
        except Exception as e:
            raise ValueError(f"Error converting {value} to a number: {e}")
        
    def value(self, items):
        if isinstance(items[0], Token):
            return items[0].value
        return items[0]

    def explicit_tuple(self, items):
        return tuple(self._extract_value(item) for item in items)

    def bool_value(self, items):
        return self._extract_value(items[0])
    
    def simple_number(self, items):
        return self._extract_value(items[0])

    
    def ordered_params(self, items):

        def dense(self, items):
            params = items[0]
            if isinstance(params, list):
                param_dict = {}
                if len(params) >= 1:
                    param_dict['units'] = self._extract_value(items[0])
                if len(params) >= 2:
                    param_dict['activation'] = self._extract_value(items[1])
                params = param_dict
            return {'type': 'Dense', 'params': params}

        def conv2d(self, items):
            params = items[0]
            if isinstance(params, list):
                param_dict = {}
                if len(params) >= 1:
                    param_dict['filters'] = params[0]
                if len(params) >= 2:
                    param_dict['kernel_size'] = params[1]
                if len(params) >= 3:
                    param_dict['activation'] = params[2]
                params = param_dict
            else:
                param_dict = params
            return {'type': 'Conv2D', 'params': params}

        def max_pooling2d(self, items):
            params = items[0]
            if isinstance(params, list):
                param_dict = {}
                if len(params) >= 1:
                    param_dict['pool_size'] = params[0]
                if len(params) >= 2:
                    param_dict['strides'] = params[1]
                if len(params) >= 3:
                    param_dict['padding'] = params[2]
                params = param_dict
            return {'type': 'MaxPooling2D', 'params': params}
            
        
    def named_kernel_size(self, items):
        return {"kernel_size": self._extract_value(items[0])}

    def named_filters(self, items):
        return {"filters": self._extract_value(items[0])}

    def named_units(self, items):  
        return {"units": self._extract_value(items[0])}

    def activation_param(self, units):
        return {"activation": self._extract_value(units)}

    def named_activation(self, items): 
        return {"activation": self._extract_value(items[0])}

    def named_strides(self, items):
        value = self._extract_value(items[0])
        return {"strides": value}

    def named_padding(self, items):
        value = self._extract_value(items[0])
        return {"padding": value}

    def named_rate(self, items):
        return {"rate": self._extract_value(items[0])}

    def named_dilation_rate(self, items):
        return {"dilation_rate": self._extract_value(items[0])}

    def named_groups(self, items):
        return {"groups": self._extract_value(items[0])}

    def named_size(self, items):
        name = str(items[0])
        value = tuple(int(x) for x in items[2].children)
        return {name: value}

    def named_dropout(self, items):
        return {"dropout": self._extract_value(items[0])}

    def named_return_sequences(self, items):
        return {"return_sequences": self._extract_value(items[0])}

    def named_num_heads(self, items):
        return {"num_heads": self._extract_value(items[0])}

    def named_ff_dim(self, items):
        return {"ff_dim": self._extract_value(items[0])}

    def named_input_dim(self, items):
        return {"input_dim": self._extract_value(items[0])}

    def named_output_dim(self, items):
        return {"output_dim": self._extract_value(items[0])}

    def groups_param(self, items):
        return {'groups': self._extract_value(items[0])}

    def device_param(self, items):
        return {'device': self._extract_value(items[0])}

    ### End named_params ################################################

    
    ### Advanced Layers #################################

    def attention(self, items):
        params = items[0] if items else None
        return {'type': 'Attention', 'params': params}

    def residual(self, items):
        params = items[0] if items else None
        return {'type': 'ResidualConnection', 'params': params}

    def inception(self, items):
        params = items[0] if items else None
        return {'type': 'Inception', 'params': params}

    def graph(self, items):
        return items[0]

    
    def graph_attention(self, items):
        params = self._extract_value(items[0])
        return {'type': 'GraphAttention', 'params': params}
    
    def dynamic(self, items):
        params = items[0] if items else None
        return {'type': 'DynamicLayer', 'params': params}
    
    def noise_layer(self, items):
        return {'type': items[0].data.capitalize(), 'params': items[0].children[0]}

    def normalization_layer(self, items):
        return {'type': items[0].data.capitalize(), 'params': items[0].children[0]}

    def regularization_layer(self, items):
        return {'type': items[0].data.capitalize(), 'params': items[0].children[0]}

    def custom_layer(self, items):
        params = self._extract_value(items[0])
        return {'type': 'Capsule', 'params': params}
    
    def capsule(self, items):
        params = items[0] if items else None
        return {'type': 'CapsuleLayer', 'params': params}

    def squeeze_excitation(self, items):
        params = self._extract_value(items[0]) if items else None
        return {'type': 'SqueezeExcitation', 'params': params}

    def quantum(self, items):
        params = items[0] if items else None
        return {'type': 'QuantumLayer', 'params': params}

    def transformer(self, items):
        params = self._extract_value(items[0])
        return {'type': 'TransformerEncoder', 'params': params}
    
    def embedding(self, items):
        params = items[0] if items else None
        return {'type': 'Embedding', 'params': params}
    
    def lambda_(self, items):
        return {'type': 'Lambda', 'params': {'function': self._extract_value(items[0])}}
    
    def add(self, items):
        return {'type': 'Add', 'params': items[0]}

    def subtract(self, items):
        return {'type': 'Subtract', 'params': items[0]}

    def multiply(self, items):
        return {'type': 'Multiply', 'params': items[0]}

    def average(self, items):
        return {'type': 'Average', 'params': items[0]}

    def maximum(self, items):
        return {'type': 'Maximum', 'params': items[0]}

    def concatenate(self, items):
        return {'type': 'Concatenate', 'params': items[0]}

    def dot(self, items):
        return {'type': 'Dot', 'params': items[0]}

    def gaussian_noise(self, items):
        return {'type': 'GaussianNoise', 'params': items[0]}

    def gaussian_dropout(self, items):
        return {'type': 'GaussianDropout', 'params': items[0]}

    def alpha_dropout(self, items):
        return {'type': 'AlphaDropout', 'params': items[0]}

    def batch_normalization(self, items):
        return {'type': 'BatchNormalization', 'params': items[0]}

    def layer_normalization(self, items):
        return {'type': 'LayerNormalization', 'params': items[0]}

    def instance_normalization(self, items):
        return {'type': 'InstanceNormalization', 'params': items[0]}

    def group_normalization(self, items):
        return {'type': 'GroupNormalization', 'params': items[0]}

    def spatial_dropout1d(self, items):
        return {'type': 'SpatialDropout1D', 'params': items[0]}

    def spatial_dropout2d(self, items):
        return {'type': 'SpatialDropout2D', 'params': items[0]}

    def spatial_dropout3d(self, items):
        return {'type': 'SpatialDropout3D', 'params': items[0]}

    def activity_regularization(self, items):
        return {'type': 'ActivityRegularization', 'params': items[0]}

    def l1_l2(self, items):
        return {'type': 'L1L2', 'params': items[0]}

    def custom(self, items):
        return {'type': items[0], 'params': items[1]}
    
    # Custom Shape Propagation
    def custom_shape(self, items):
    # Example: CustomShape(MyLayer, (32, 32))
        layer_name = self._extract_value(items[0])
        custom_dims = self._extract_value(items[1])
        return {"type": "CustomShape", "layer": layer_name, "custom_dims": custom_dims}

### Automatic Hyperparameters Optimization ###

    def hpo_expr(self, items):
        return {"hpo": items[0]}

    def hpo_choice(self, items):
        return {"type": "categorical", "values": [self._extract_value(x) for x in items]}

    def hpo_range(self, items):
        start = self._extract_value(items[0])
        end = self._extract_value(items[1])
        step = self._extract_value(items[2]) if len(items) > 2 else 1
        return {"type": "range", "start": start, "end": end, "step": step}

    def hpo_log_range(self, items):
        low = self._extract_value(items[0])
        high = self._extract_value(items[1])
        return {"type": "log_range", "low": low, "high": high}

### Neural Architecture Search ###

    def layer_choice(self, items):
        return {"hpo_type": "layer_choice", "options": items}


### Frameworks Detection ###
    def parse_network(self, config: str, framework: str = 'auto'):
        tree = network_parser.parse(config)
        model = self.transform(tree)
        
        # Auto-detect framework
        if framework == 'auto':
            framework = self._detect_framework(model)
            
        # Add framework-specific metadata
        model['framework'] = framework
        model['shape_info'] = []
        
        return model

    def _detect_framework(self, model):
        # Detect framework based on layer types and parameters
        for layer in model['layers']:
            if 'torch' in layer.get('params', {}).values():
                return 'pytorch'
            if 'keras' in layer.get('params', {}).values():
                return 'tensorflow'
        return 'tensorflow'  # default





import torch
import tensorrt as trt

def get_device(preferred_device="auto"):
    """ Selects the best available device: GPU, CPU, or future accelerators """
    if preferred_device.lower() == "gpu" and torch.cuda.is_available():
        return torch.device("cuda")
    elif preferred_device.lower() == "cpu":
        return torch.device("cpu")
    elif torch.cuda.is_available():
        return torch.device("cuda")
    else:
        return torch.device("cpu")

def run_inference(model, data, execution_config):
    """ Runs inference on the specified device """
    device = get_device(execution_config.get("device", "auto"))
    model.to(device)
    data = data.to(device)

    with torch.no_grad():
        output = model(data)

    return output.cpu()

def optimize_model_with_tensorrt(model):
    """ Converts model to TensorRT for optimized inference """
    model.eval()
    device = get_device("gpu")

    # Dummy input for tracing
    dummy_input = torch.randn(1, *model.input_shape).to(device)

    traced_model = torch.jit.trace(model, dummy_input)
    trt_model = torch.jit.freeze(traced_model)

    return trt_model

def run_inference(model, data, execution_config):
    """ Runs optimized inference using TensorRT or PyTorch """
    device = get_device(execution_config.get("device", "auto"))
    
    if device.type == "cuda":
        model = optimize_model_with_tensorrt(model)

    model.to(device)
    data = data.to(device)

    with torch.no_grad():
        output = model(data)

    return output.cpu()


import logging
import json
import time
import torch
import numpy as np
import psutil
import plotly.graph_objects as go
from graphviz import Digraph
from typing import Dict, Tuple, Optional, Any, List

import sys
import os

# Add the parent directory of 'neural' to sys.path
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

from parser.parser import ModelTransformer
from pretrained_models.pretrained import PretrainedModelHub

class PerformanceMonitor:
    def __init__(self):
        self.resource_history = []

    def monitor_resources(self):
        """Monitor CPU, memory, and GPU usage."""
        cpu_usage = psutil.cpu_percent()
        memory_usage = psutil.virtual_memory().percent
        gpu_memory = 0
        if torch.cuda.is_available():
            gpu_memory = torch.cuda.memory_allocated() / (1024 ** 3)  # GB
        io_counters = psutil.disk_io_counters()
        io_usage = (io_counters.read_bytes + io_counters.write_bytes) / (1024 ** 2)  # MB

        self.resource_history.append({
            "timestamp": time.time(),
            "cpu_usage": cpu_usage,
            "memory_usage": memory_usage,
            "gpu_memory": gpu_memory,
            "io_usage": io_usage
        })
        return self.resource_history[-1]


class ShapePropagator:
    
    def __init__(self, debug=False):
        self.debug = debug
        self.shape_history = []
        self.layer_connections = []
        self.current_layer = 0
        self.execution_trace = []  # Stores nntrace logs
        self.performance_monitor = PerformanceMonitor()
        self.hub = PretrainedModelHub()
        
        # Framework compatibility mappings
        self.param_aliases = {
            'Conv2D': {'filters': 'out_channels', 'kernel_size': 'kernel_size'},
            'BatchNormalization': {'axis': 'dim'},
            'Dense': {'units': 'out_features'},
            'LSTM': {'units': 'hidden_size'},
            'BatchNormalization': {'momentum': 'decay'}
        }
        
        # Initialize visualization
        self.dot = Digraph(comment='Neural Network Architecture')
        self.dot.attr('node', shape='record', style='filled', fillcolor='lightgrey')

    def propagate(self, input_shape: Tuple[Optional[int], ...], 
              layer: Dict[str, Any], 
              framework: str = 'tensorflow') -> Tuple[Optional[int], ...]:
        """Processes a layer and logs shape changes for nntrace."""
        layer_type = layer["type"]
        params = layer.get("params", {})
        kernel_size = tuple(params.get("kernel_size", [3, 3]) if isinstance(params.get("kernel_size"), list) else (3, 3))
        params["kernel_size"] = kernel_size  # Ensure tuple in params

        start_time = time.time()  # Measure execution time

        output_shape = self._process_layer(input_shape, layer, framework)
        prev_layer = self.current_layer - 1 if self.current_layer > 0 else None

        # Compute FLOPs, memory, compute_time, and transfer_time
        flops, mem_usage, compute_time, transfer_time = self._compute_performance(layer, input_shape, output_shape)

        # Capture nntrace log with additional timing details
        trace_entry = {
            "layer": layer_type,
            "input_shape": input_shape,
            "output_shape": output_shape,
            "flops": flops,
            "memory": mem_usage,
            "execution_time": time.time() - start_time,
            "compute_time": compute_time,
            "transfer_time": transfer_time,
        }
        self.execution_trace.append(trace_entry)

        resources = self.performance_monitor.monitor_resources()
        trace_entry.update({
            "cpu_usage": resources["cpu_usage"],
            "memory_usage": resources["memory_usage"],
            "gpu_memory": resources["gpu_memory"],
            "io_usage": resources["io_usage"]
        })

        if self.debug:
            print(f"TRACE: {trace_entry}")  # Debugging output

        self._visualize_layer(layer['type'], output_shape)  # Creates node and increments self.current_layer
        if prev_layer is not None:
            self._create_connection(prev_layer, self.current_layer - 1)  # Connect previous to current
        return output_shape

###############################
### Performance Computation ###
###############################

    def _compute_performance(self, layer: dict, input_shape: tuple, output_shape: tuple) -> tuple:
        """Compute performance metrics (FLOPs, memory usage, etc.)."""
        # Replace None with 1 to avoid NoneType math errors
        input_shape = tuple(1 if dim is None else dim for dim in input_shape)
        output_shape = tuple(1 if dim is None else dim for dim in output_shape)

        # FLOPs calculation (example for Conv2D)
        if layer['type'] == 'Conv2D':
            kernel_size = layer['params']['kernel_size']
            filters = layer['params']['filters']
            flops = np.prod(kernel_size) * np.prod(output_shape) * input_shape[-1]
        else:
            flops = 0  # Default for other layers

        # Memory usage (output tensor size in MB)
        memory_usage = np.prod(output_shape) * 4 / (1024 ** 2)  # 4 bytes per float

        # Simplified timing estimates
        compute_time = flops / 1e9  # 1 GFLOP/s
        transfer_time = memory_usage * 1e3 / 1e9  # 1 GB/s

        return flops, memory_usage, compute_time, transfer_time

##################################################
### Send execution trace data to the dashboard ###
##################################################
    def get_trace(self):
        trace = []
        for layer_type, exec_time, comp_time, trans_time, params, flops, memory, grad_norm, dead_ratio, mean_act, anomaly in self.execution_trace:
            kernel_size = params.get("kernel_size", (1, 1))
            if isinstance(kernel_size, list):
                print(f"WARNING: Converting list kernel_size {kernel_size} to tuple for {layer_type}")
                kernel_size = tuple(kernel_size)
            elif not isinstance(kernel_size, tuple):
                print(f"WARNING: Unexpected kernel_size type {type(kernel_size)} for {layer_type}, defaulting to (1, 1)")
                kernel_size = (1, 1)
            trace.append({
                "layer": layer_type, "execution_time": exec_time, "compute_time": comp_time,
                "transfer_time": trans_time, "kernel_size": kernel_size,
                "flops": flops, "memory": memory, "grad_norm": grad_norm, "dead_ratio": dead_ratio,
                "mean_activation": mean_act, "anomaly": anomaly
            })
            
        return trace

    def _process_layer(self, input_shape, layer, framework):
        layer_type = layer['type']
        params = self._standardize_params(layer.get('params', {}), layer_type, framework)
        # Unified parameter handling
        handler_name = f"_handle_{layer_type.lower()}"
        if hasattr(self, handler_name):
            output_shape = getattr(self, handler_name)(input_shape, params)
        else:
            output_shape = self._handle_default(input_shape, params)
        return output_shape

    def _standardize_params(self, params, layer_type, framework):
        # Convert framework-specific parameters to canonical form
        standardized = {}
        aliases = self.param_aliases.get(layer_type, {})
        
        for k, v in params.items():
            if framework == 'pytorch' and k in aliases.values():
                standardized[aliases[k]] = v
            else:
                standardized[k] = v
                
        # Add framework-specific defaults
        if framework == 'pytorch':
            standardized.setdefault('data_format', 'channels_first')
        else:
            standardized.setdefault('data_format', 'channels_last')
            
        return standardized

####################################################################
### Shape propagation through 2 Dimensional Convolutional Layers ###
####################################################################

    def _handle_conv2d(self, input_shape, params):
        """Calculates the output shape for a Conv2D layer.

        This method computes the output shape based on the input shape,
        kernel size, stride, padding, and data format.

        Args:
            input_shape (tuple): Input tensor shape.
            params (dict): Layer parameters.

        Returns:
            tuple: Output tensor shape.
        """    
        data_format = params['data_format']
        if data_format == 'channels_first':
            spatial_dims = input_shape[2:]  # (height, width)
        else:  # channels_last
            spatial_dims = input_shape[1:3]  # (height, width)

        # Handle kernel_size as tuple or integer
        kernel = params['kernel_size']
        if isinstance(kernel, int):
            kernel = (kernel,) * len(spatial_dims)
        elif isinstance(kernel, tuple):
            if len(kernel) != len(spatial_dims):
                raise ValueError("Kernel size must match spatial dimensions")

        kernel_size = tuple(params.get("kernel_size", [3, 3]) if isinstance(params.get("kernel_size"), list) else (3, 3))

        stride = params.get('stride', 1)
        padding = self._calculate_padding(params, input_shape[1] if data_format=='channels_last' else input_shape[2])

        # Ensure padding is a tuple matching spatial dimensions
        if isinstance(padding, int):
            padding = (padding,) * len(kernel)
        elif isinstance(padding, (list, tuple)):
            if len(padding) != len(kernel):
                raise ValueError(f"Padding length {len(padding)} != kernel dimensions {len(kernel)}")
            padding = tuple(padding)
        else:
            raise TypeError(f"Invalid padding type: {type(padding)}")

        # Calculate output shape per spatial dimension
        output_spatial = [
            (dim + 2*pad - k) // stride + 1
            for dim, k, pad in zip(spatial_dims, kernel, padding)
        ]

        # Reconstruct full shape
        if data_format == 'channels_first':
            return (input_shape[0], params['filters'], *output_spatial)
        else:
            return (input_shape[0], *output_spatial, params['filters'])
        
    def _handle_maxpooling2d(self, input_shape, params):
        data_format = params.get('data_format', 'channels_last')
        pool_size = params['pool_size']
        stride = params.get('stride', pool_size)

        # Handle stride as tuple or integer
        if isinstance(stride, (tuple, list)):
            stride_h, stride_w = stride
        else:
            stride_h = stride_w = stride

        # Calculate spatial dimensions based on data format
        if data_format == 'channels_last':
            # TensorFlow: input_shape = (batch, height, width, channels)
            new_height = input_shape[1] // stride_h
            new_width = input_shape[2] // stride_w
            return (input_shape[0], new_height, new_width, input_shape[3])
        else:
            # PyTorch: input_shape = (batch, channels, height, width)
            new_height = input_shape[2] // stride_h
            new_width = input_shape[3] // stride_w
            return (input_shape[0], input_shape[1], new_height, new_width)
    
    def _handle_flatten(self, input_shape, params):
        # If there is a batch dimension, keep it.
        if len(input_shape) >= 1:
            batch = input_shape[0]
            # Multiply all dimensions after the batch dimension
            flattened = np.prod(input_shape[1:])
            return (batch, flattened)
        else:
            return (np.prod(input_shape),)


    def _handle_dense(self, input_shape, params):
        # If input_shape has two or more dimensions, preserve the batch dimension.
        if len(input_shape) >= 2:
            return (input_shape[0], params['units'])
        else:
            return (params['units'],)

    def _handle_output(self, input_shape, params):
        # Preserves the batch dimension and converts the feature dimension to the number of output units.
        if len(input_shape) >= 2:
            return (input_shape[0], params['units'])
        else:
            return (params['units'],)


    # Handle default helper
    def _handle_default(self, input_shape, params):
        # Default handler for unsupported layers
        return input_shape
    
    ### Padding detection, extraction and calculation ###
    def _calculate_padding(self, params, input_dim):
        """Calculates padding based on provided parameters and input dimension.

        This method handles different padding types: integer, list, or string.
        It returns the appropriate padding value based on the input.

        Args:
            params (dict): Layer parameters containing padding information.
            input_dim (int): Input dimension.

        Returns:
            int or tuple or list: Calculated padding value.
        """
        padding = params.get('padding', 0)
    
        if isinstance(padding, int):
            return padding
        elif isinstance(padding, (list, tuple)):
            return tuple(padding)
        elif padding == 'same':
            # Handle kernel_size as tuple or integer
            kernel = params['kernel_size']
            if isinstance(kernel, int):
                return (kernel - 1) // 2
            elif isinstance(kernel, tuple):
                return tuple((k - 1) // 2 for k in kernel)  # Process each dimension
        elif padding == 'valid':
            return 0
        else:
            return [padding] * (input_dim - 2)
    
    ### Layers Shape Propagation Visualization ###
    def _visualize_layer(self, layer_name, shape):
        label = f"{layer_name}\n{shape}"
        self.dot.node(str(self.current_layer), label)
        self.shape_history.append((layer_name, shape))
        self.current_layer += 1

    def _create_connection(self, from_id, to_id):
        self.layer_connections.append((from_id, to_id))
        self.dot.edge(str(from_id), str(to_id))

    def generate_report(self):
        """Generate interactive visualization and shape report"""
        # Plotly visualization
        fig = go.Figure()
        
        # Add shape dimensions as bar chart
        shapes = [str(s[1]) for s in self.shape_history]
        fig.add_trace(go.Bar(
            x=[s[0] for s in self.shape_history],
            y=[np.prod(s[1]) for s in self.shape_history],
            text=shapes,
            name='Parameter Count'
        ))
        
        fig.update_layout(
            title='Network Shape Propagation',
            xaxis_title='Layer',
            yaxis_title='Parameters',
            template='plotly_white'
        )
        
        return {
            'dot_graph': self.dot,
            'plotly_chart': fig,
            'shape_history': self.shape_history
        }

    def _log_shape(self, shape, stage):
        if self.debug:
            logging.info(f"{stage.upper()} SHAPE: {shape}")
            logging.debug(f"Shape details: {self._shape_analysis(shape)}")

    def _shape_analysis(self, shape):
        return {
            'total_parameters': np.prod([d for d in shape if d]),
            'spatial_dims': shape[2:-1] if len(shape) > 2 else None,
            'channel_dim': shape[1] if len(shape) > 1 else None
        }
    
    ### Loading Pretrained Models ####

    def load_pretrained(self, model_name, pretrained=True):
        model = self.hub.load(model_name, pretrained)
        # Propagate shapes through pretrained model
        input_shape = (1, 3, 224, 224)  # Default for ResNet50
        for layer in model.layers:
            input_shape = self.propagate(input_shape, layer, "pytorch")

### Shape Validation for Error Handling ###

class ShapeValidator:
    @staticmethod
    def validate_layer(layer_type, input_shape, params):
        validators = {
            'Conv2D': lambda: ShapeValidator._validate_conv(input_shape, params),
            'Dense': lambda: ShapeValidator._validate_dense(input_shape, params)
        }
        
        if validator := validators.get(layer_type):
            validator()
            
    @staticmethod
    def _validate_conv(input_shape, params):
        if len(input_shape) != 4:
            raise ValueError(f"Conv layers need 4D input. Got {len(input_shape)}D")
        if params['kernel_size'] > input_shape[2]:
            raise ValueError(f"Kernel size {params['kernel_size']} "
                           f"exceeds input dimension {input_shape[2]}")
        
    @staticmethod
    def _validate_dense(input_shape, params):
        if len(input_shape) > 2:
            raise ValueError(
                f"Dense layer expects 2D input (batch, features). "
                f"Got {len(input_shape)}D: {input_shape}"
            )
# Unified parameter handling for TF/PyTorch
FRAMEWORK_DEFAULTS = {
    'tensorflow': {
        'data_format': 'channels_last',
        'padding': 'same'
    },
    'pytorch': {
        'data_format': 'channels_first',
        'padding': 0
    }
}

def get_framework_params(framework):
    return FRAMEWORK_DEFAULTS.get(framework.lower(), FRAMEWORK_DEFAULTS['tensorflow'])

### Real-Time Shape Visualization ###
def get_shape_data(self):
        """Returns shape history as JSON."""
        return json.dumps([
        {"layer": layer[0], "output_shape": layer[1]} 
        for layer in self.shape_history
    ])

def _calculate_shape(self, input_shape, layer):
    if layer["type"] == "Dense":
        return (input_shape[0], layer["params"]["units"])
    elif layer["type"] == "Conv2D":
        return (input_shape[0], input_shape[1], input_shape[2], layer["params"]["filters"])
    elif layer["type"] == "Flatten":
        return (input_shape[0], np.prod(input_shape[1:]))
    return input_shape

### Compute FLOPs and memory usage for visualization ###
def compute_flops_params(layer, input_shape):
    """Estimate FLOPs and parameter counts for a given layer."""
    if layer["type"] == "Dense":
        units = layer["params"]["units"]
        params = input_shape[1] * units + units  # Weights + biases
        flops = 2 * params  # Two operations per weight (multiply + add)
    
    elif layer["type"] == "Conv2D":
        filters = layer["params"]["filters"]
        kernel_size = layer["params"]["kernel_size"]
        stride = layer["params"].get("stride", 1)
        params = (kernel_size[0] * kernel_size[1] * input_shape[-1] + 1) * filters
        output_height = (input_shape[1] - kernel_size[0]) // stride + 1
        output_width = (input_shape[2] - kernel_size[1]) // stride + 1
        flops = params * output_height * output_width
    
    return params, flops

#######################################
### Gradient Flow Visualization #######
#######################################
def register_gradient_hooks(model):
    """Attaches hooks to capture gradient magnitudes during backprop."""
    gradient_trace = []

    def hook(module, grad_input, grad_output):
        if grad_output[0] is not None:
            grad_norm = grad_output[0].detach().abs().mean().item()
            gradient_trace.append({"layer": module.__class__.__name__, "grad_norm": grad_norm})

    for layer in model.children():
        layer.register_backward_hook(hook)

    return gradient_trace

#####################################
### Dead Neurons Detection ##########
#####################################
def detect_dead_neurons(layer, input, output):
    """Detects inactive neurons (dead neurons)."""
    dead_neurons = (output.detach() == 0).sum().item()
    total_neurons = output.numel()
    dead_ratio = dead_neurons / total_neurons

    return {"layer": layer.__class__.__name__, "dead_ratio": dead_ratio}

######################################
### Activation Anomalies Detection ###
######################################
def detect_activation_anomalies(layer, input, output):
    """Flags NaNs, extremely high activations, or overflows."""
    mean_activation = output.detach().abs().mean().item()
    has_nan = torch.isnan(output).sum().item() > 0
    is_exploding = mean_activation > 1000  # Arbitrary threshold for huge activations

    return {
        "layer": layer.__class__.__name__,
        "mean_activation": mean_activation,
        "anomaly": has_nan or is_exploding
    }


######################
### Step Debugging ###
######################
def step_debug_hook(module, input, output):
    """Pauses execution at this layer for manual debugging."""
    print(f"Paused at layer: {module.__class__.__name__}")
    print(f"Input shape: {input[0].shape}, Output shape: {output.shape}")

    # Wait for user input before continuing
    input("Press Enter to continue...")



import os
import sys

# Add the parent directory to the Python path
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))


from neural.shape_propagation.shape_propagator import ShapePropagator
import pytest



def test_invalid_shape():
    input_shape = (28, 28)
    layers = [{"type": "Conv2D", "filters": 32, "kernel_size": (3, 3)}]
    with pytest.raises(ValueError):
        propagate_shape = ShapePropagator.propagate()
        propagate_shape(input_shape, layers[0])